#include "stdafx.h"
#include "MgaXslt.h"
#include "CommonStl.h"

#include <xalanc/Include/PlatformDefinitions.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xalanc/XalanTransformer/XalanTransformer.hpp>


STDMETHODIMP CMgaXslt::ApplyXslt( BSTR p_stylesheet_f_name, BSTR p_in_f_name, BSTR p_out_f_name, BSTR * out_error)
{
	XALAN_USING_XERCES(XMLPlatformUtils)
	XALAN_USING_XERCES(XMLException)

	XALAN_USING_XALAN(XalanTransformer)
	XALAN_USING_XALAN(XalanCompiledStylesheet)

	std::string style_sheet( 1 + GetCharLength( p_stylesheet_f_name), ' ');
	std::string in_file( 1 + GetCharLength( p_in_f_name), ' ');
	std::string out_file( 1 + GetCharLength( p_out_f_name), ' ');
	std::string error;

	CopyTo( p_stylesheet_f_name, style_sheet);
	CopyTo( p_in_f_name, in_file);
	CopyTo( p_out_f_name, out_file);

	int	theResult = 0;

	// Call the static initializer for Xerces.
	try
	{
		XMLPlatformUtils::Initialize();
	}
	catch (const XMLException&)
	{
		theResult = E_XSLT_XERCES_INIT;
	}

	try 
	{
		if ( theResult == 0)
		{
			// Initialize Xalan.
			XalanTransformer::initialize();
			
			// Create a XalanTransformer.
			XalanTransformer theXalanTransformer; 
			
			// Compile the stylesheet.
			const XalanCompiledStylesheet*	theCompiledStylesheet = 0;
			
			theResult =	theXalanTransformer.compileStylesheet( style_sheet.c_str(), theCompiledStylesheet);
			
			if (theResult != 0)
			{
				theResult = E_XSLT_COMPILE_ERROR;
				error = theXalanTransformer.getLastError();
			}
			else
			{
				assert(theCompiledStylesheet != 0);
		
				// Do the transform.
				theResult = theXalanTransformer.transform
					( in_file.c_str()
					, theCompiledStylesheet
					, out_file.c_str()
					);
				
				// I don't have to do this, since the transformer will
				// clean this up when it's destroyed, but why not...
				theXalanTransformer.destroyStylesheet(theCompiledStylesheet);

				if ( theResult != 0)
				{
					theResult = E_XSLT_TRANSF_ERROR;
					error = theXalanTransformer.getLastError();
				}
			}
		}
	} 
	catch( ...)
	{
		theResult = E_XSLT_ERROR;
	}

	// Terminate Xalan...
	XalanTransformer::terminate();

	// Terminate Xerces...
	XMLPlatformUtils::Terminate();

	// Clean up the ICU, if it's integrated.
	XalanTransformer::ICUCleanUp();

	CopyTo( error, out_error);

	return (theResult == 0)? S_OK : theResult;
}
