#!/usr/bin/python
# Copyright (c) 2004 ISIS, Vanderbilt Univeristy 
#
# Authors: Gyorgy Balogh, Peter Volgyesi
#

""" GME Build System

This file contains useful tools for building GME

"""

import os
import os.path
import win32com.client

def runExe( exePath, commandLineParams ):
    """
    Runs an external exe
    params
        exePath             : full path of the executable
        commandLineParams   : command line parameters
    """
    ret = os.spawnl( os.P_WAIT, exePath, exePath, commandLineParams )
    if ret != 0:
        raise "Execution of " + exePath + " failed. Exit code: %d" % (ret,)

def system( command ):
    """
    Executes a system command (throws an exception on error)
    params
        command : command to execute
    """
    ret = os.system( command )
    if ret != 0:
        raise Exception, command + " failed. Exit code: %d" % (ret,)

def buildMsDevProject( dsw_path, config_name ):
    """
    Builds an msdev project
    params
       dsw_path     : full path of the dsw file
       config_name  : name of the configuration containing the project name (e.g.: "GME - Win32 Release")
    """
    print "Compiling " + config_name + "..."
    app = win32com.client.Dispatch("MSDev.Application")
    app.Visible = True
    app.Documents.Open( dsw_path )
    found = 0
    for proj in app.Projects:
        for conf in proj.Configurations:
            if conf.Name == config_name:
                app.Clean( conf )
                app.Build( conf )
                if app.Errors > 0:                    
                    print "There are ", app.Errors, " errors."
                    raise Exception, "Compile error"
                if app.Warnings > 0:
                    print "There are ", app.Errors, " warnings."
                found = 1
    app.Quit()
    if found == 0:
        raise Exception, "\"" + config_name + "\" configuration not found in \"" + dsw_path + "\""

def xme2mga( xmlFile, mgaFile, paradigm ):
    """
    Generates .mga file from an .xme file
    """
    mgaParser  = win32com.client.Dispatch( "MGA.MgaParser" )
    mgaProject = win32com.client.Dispatch( "MGA.MgaProject" )
    mgaProject.Create( "MGA="+mgaFile, paradigm )
    mgaParser.ParseProject( mgaProject, xmlFile )
    mgaProject.Close()    
    

def registerParadigm( xmpFile, paradigm ):
    """    
    """
    # Get modules for python MgaRegistar interface definitions
        
    mgaReg = win32com.client.Dispatch( "MGA.MgaRegistrar" )
    mgaReg.RegisterParadigmFromData( "XML=" + xmpFile, paradigm, 2 )
    
    return mgaReg.ParadigmGUIDString(2, paradigm)
    
def buildISProject(ISProject, config, release, properties):
    print "Building InstallShield - " + ISProject + " - " + config + " - " + release
    ISWiProject = win32com.client.Dispatch( "ISWiAutomation.ISWiProject" )
    ISWiProject.ImportProject(ISProject + ".ism", ISProject + ".isv")
    ISWiProject.CloseProject()
    
    ISWiProject.OpenProject(ISProject + ".ism")
    for property in properties.keys():
        ISProperty = ISWiProject.ISWiProperties.Item(property)
        ISProperty.Value = properties[property]
        print "\tSet property " + ISProperty.Name + " = " + ISProperty.Value
        
    ISConfig = ISWiProject.ISWiProductConfigs.Item(config)
    ISRelease = ISConfig.ISWiReleases.Item(release)
    ISRelease.Build()
    ISWiProject.CloseProject()
    print "Errors: " + str(ISRelease.BuildErrorCount)
    print "Warnings: " + str(ISRelease.BuildWarningCount)
    
    if ISRelease.BuildErrorCount > 0:
        raise "InstallShield error(s) occured"
        
    print "OK"
        
# custom registry stuff

"""
def warning( message ):
    print "warning:" + message
    
def error( message ):
    print "error:" + message

def message( message ):
    print message
    """
