#ifndef META_INTERPRETER_H
#define META_INTERPRETER_H

#include <fstream.h>
#include "Builder.h"
#include "pragma.h"

class CParadigmBuilder;

class CUMLClassBase;

class CFolderBase;
class CFolderBuilder;
class CFolderProxy;

class CAttributeBuilder;

class CObjectClassBase;
class CFCOClassBase;
class CFCOBuilder;
class CModelBuilder;
class CAtomBuilder;
class CReferenceBuilder;
class CSetBuilder;
class CConnectionBuilder;
class CObjectProxyBuilder;

class CConstraintBuilder;
class CConstraintFuncBuilder;



class CAspect;
class CRole;
class CPart;
class CPointerItem;

#define STR_MAX 256

/*
#define IFLAG_IMPLEMENTATION	0x1
#define IFLAG_INTERFACE			(0x1 << 1)
#define IFLAG_FULL				(IFLAG_IMPLEMENTATION | IFLAG_INTERFACE)
*/


enum inheritance_type {full, interf, implement};

typedef CTypedPtrList<CPtrList, CUMLClassBase*> CUMLClassBaseList;

typedef CTypedPtrList<CPtrList, CFolderBase*> CFolderBaseList;
typedef CTypedPtrList<CPtrList, CFolderBuilder*> CFolderBuilderList;
typedef CTypedPtrList<CPtrList, CFolderProxy*> CFolderProxyList;

typedef CTypedPtrList<CPtrList, CAttributeBuilder*> CAttributeBuilderList;

typedef CTypedPtrList<CPtrList, CObjectClassBase*> CObjectClassBaseList;
typedef CTypedPtrList<CPtrList, CFCOClassBase*> CFCOClassBaseList;
typedef CTypedPtrList<CPtrList, CFCOBuilder*> CFCOBuilderList;
typedef CTypedPtrList<CPtrList, CModelBuilder*> CModelBuilderList;
typedef CTypedPtrList<CPtrList, CAtomBuilder*> CAtomBuilderList;
typedef CTypedPtrList<CPtrList, CReferenceBuilder*> CReferenceBuilderList;
typedef CTypedPtrList<CPtrList, CSetBuilder*> CSetBuilderList;
typedef CTypedPtrList<CPtrList, CConnectionBuilder*> CConnectionBuilderList;
typedef CTypedPtrList<CPtrList, CObjectProxyBuilder*> CObjectProxyBuilderList;

typedef CTypedPtrList<CPtrList, CConstraintBuilder*> CConstraintBuilderList;
typedef CTypedPtrList<CPtrList, CConstraintFuncBuilder*> CConstraintFuncBuilderList;


typedef CTypedPtrList<CPtrList, CRole*> CRoleList;
typedef CTypedPtrList<CPtrList, CPart*> CPartList;
typedef CTypedPtrList<CPtrList, CAspect*> CAspectList;
typedef CTypedPtrList<CPtrList, CPointerItem*> CPointerItemList;


// the two SortList helper functions aid in ensuring that
// the Metamodel generation is deterministic by sorting a list
// alphabetically.

template <class A>
void SortList( const A * listToSort,
					 A &dstList )
{
	if( listToSort )
	{
		dstList.RemoveAll( );
		POSITION pos = listToSort->GetHeadPosition( );
		while( pos )
		{
			dstList.AddTail( listToSort->GetNext( pos ) );
		}
		SortList( dstList );
	}
	return;
}


template<class BASE_CLASS, class TYPE>
void SortList(CTypedPtrList<BASE_CLASS, TYPE> &listToSort)
{
	if( !listToSort.IsEmpty( ) )
	{
		// now, sort listToSort alphabetically
		int count = listToSort.GetCount( );
		for( int i=0; i < count; i++ )
		{
			for( int j=0; j < count - i - 1; j++ )
			{
				POSITION pos = listToSort.FindIndex( j );
				POSITION pos2 = listToSort.FindIndex( j+1 );
				if( pos && pos2 )
				{
					TYPE obj = listToSort.GetAt( pos );
					TYPE obj2 = listToSort.GetAt( pos2 );
					if( ( obj && obj2 ) &&
						( strcmp( obj->GetName( ), obj2->GetName( ) ) > 0 ) )
					{
						listToSort.SetAt(pos, obj2);
						listToSort.SetAt(pos2, obj);
					}
				}
			}
		}
	}
}



class CMetaRefBroker {
public:
	class CMetaRefNode {
	public:
		CMetaRefNode(const CBuilderObject *o = NULL, const CString &p = CString(""), metaref_type m = INVALID_METAREF) : obj(o), path(p), metaref(m) {};
		const CBuilderObject *obj;
		CString	path;
		metaref_type metaref;
	};
public:
	CMetaRefBroker();
	void	InitFromObj(const CBuilderObject *obj);
	void	CheckDulicates(void);
	metaref_type GetMetaRef(const CBuilderObject *obj = NULL, const CString &extraPath = CString(""));
	metaref_type GetRootFolderMetaRef(void) {return INITIAL_METAREF;}
	const static metaref_type INVALID_METAREF;

protected:
	void    InitFromRegistry(const CBuilderObject *obj, CComPtr<IMgaRegNode> regNode);

	const static metaref_type INITIAL_METAREF;	// Compatibility with the implied method
	metaref_type	firstfree;
	CList<CMetaRefNode, CMetaRefNode&>	metaRefDB;

};




class CParadigmBuilder
{
public:
	CParadigmBuilder()	: lastConstraintID(0) {}
	virtual ~CParadigmBuilder();
protected:
	CBuilderModelList	sheets;
	CFolderBuilderList folders;
	CFCOClassBaseList fcoClasses;		// contains all fcos, models, ... ,connections, but proxies
	CFCOBuilderList fcos;
	CModelBuilderList models;
	CAtomBuilderList atoms;
	CReferenceBuilderList refs;
	CSetBuilderList sets;
	CConnectionBuilderList conns;
	CAttributeBuilderList attrs;
	CAspectList aspects;
	CConstraintBuilderList allConstraints;
	CConstraintBuilderList constraints;
	CConstraintFuncBuilderList constraintfuncs;
	CString name;
	int	lastConstraintID;

public:
	bool CheckDuplicateNames();
	void DiscardFCOClass(CFCOClassBase *classToDrop);
	void DiscardFolder(CFolderBuilder *folderToDrop);
	void DiscardAspect(CAspect *aspectToDrop);
	void GatherSheets(CBuilderFolder *folder);
	bool Build(CBuilderFolder *rootoftheParadigm, CBuilder *builder);
	void BuildLists();
	void GatherProxies();
	void FindEquivObjects();
	void BuildFolders();
	void BuildInheritance();
	bool CheckInheritance();
	void BuildComposition();
	void BuildConnections();
	void BuildReferences();
	void BuildSets();
	void BuildAttributes();
	void BuildAspects();
	void BuildConstraints();
	void InitMetaRefs();
	void AddConstraint(CConstraintBuilder *constraint);
	CAspect *FindAspect(CBuilderObject *cond);
	void SetFlags(bool val);
	void DumpFolders(ofstream xml);
	bool DumpXML(CString filename);
	int  GetNextConstraintID()		{ return lastConstraintID++; }
	const CConnectionBuilderList &GetConnections()	{ return conns; }
};

class CUMLClassBase
{
public:
	virtual void DumpXML(ofstream xml,int depth) {}
};


class CConstraintBuilder : public CUMLClassBase, public CBuilderAtom
{
	DECLARE_CUSTOMATOM(CConstraintBuilder, CBuilderAtom)
protected:
	CString desc;
	CString expr;
	int priority;
	CString eventDepth;
	unsigned int eventMask;
public:
	void GetEventAttribute(CString eventName,unsigned int eventFlag);
	void Attach();
	virtual void Initialize();
	virtual void DumpXML(ofstream xml,int depth);
};


class CConstraintFuncBuilder : public CUMLClassBase, public CBuilderAtom
{
	DECLARE_CUSTOMATOM(CConstraintFuncBuilder, CBuilderAtom)
protected:
	CString paramlist;
	CString definition;
	CString returntype;
	CString contexttype;
	CString stereotype;
public:
	virtual void Initialize();
	virtual void DumpXML(ofstream xml,int depth);
};


///////////////////////////////////////////////////////////////////
// Folder stuff

class CFolderBase : public CUMLClassBase
{
public:
	CBuilderObject *builder;
	CFolderBuilder	*real;

	virtual void Initialize();

};

class CFolderProxy : public CFolderBase, public CBuilderAtomReference
{
	DECLARE_CUSTOMATOMREF(CFolderProxy, CBuilderAtomReference)

public:
	virtual void Initialize();
};


class CFolderBuilder : public CFolderBase, public CBuilderAtom
{
	DECLARE_CUSTOMATOM(CFolderBuilder, CBuilderAtom)
protected:
	CFolderBuilderList subfolders;
	CFCOClassBaseList fcos;
	bool root;
	CString dispName;
public:
	CFolderBaseList	parts;

public:
	virtual void Initialize();
	bool IsRoot()	{ return root; }
	void BuildComposition();
	void FindSameFolders();
	void FindSameFolders(CBuilderObjectList &diamonds, bool leftIsProxy);
	void JoinSameFolders(CBuilderObjectList &sames, bool leftIsProxy, CString commonName);
	virtual void DumpXML(ofstream xml,int depth);
	void GatherProxies();
	bool IsEmpty();

};


///////////////////////////////////////////////////////////////////
// Object stuff

class CObjectClassBase : public CUMLClassBase
{
public:
		bool flag;
		bool abstract;
		CBuilderObject* builder;
		CFCOClassBase	*real;
public:
		void Initialize();
		bool IsAbstract()			{ return abstract; }
		bool GetFlag()				{ return flag; }
		void SetFlag(bool val)		{ flag = val; }
};

class CObjectProxyBuilder : public CObjectClassBase, public CBuilderAtomReference
{
	DECLARE_CUSTOMATOMREF(CObjectProxyBuilder, CBuilderAtomReference);

public:
	virtual void Initialize();
};


class CFCOClassBase : public CObjectClassBase, public CBuilderAtom
{
	DECLARE_CUSTOMATOMBASE(CFCOClassBase, CBuilderAtom);

protected:
	CFCOClassBaseList baseClasses;
	CFCOClassBaseList subClasses;

	CFCOClassBaseList if_baseClasses;
	CFCOClassBaseList if_subClasses;

	CFCOClassBaseList imp_baseClasses;
	CFCOClassBaseList imp_subClasses;
	int namePos;
	CString dispName;
	CRoleList roles;
	CAttributeBuilderList attrs;
	CConstraintBuilderList constraints;
	CObjectClassBaseList	parts;
	bool root;

public:
	virtual void Initialize();
	void GatherProxies();
	int GetNamePosition()		{ return namePos; }
	CAttributeBuilderList& GetAttributes()	{return attrs;}
	void AddRole(CRole *role);
	CRoleList &GetRoles()		{ return roles; }
	bool CheckInheritance(CStringList &trace);
	void RegisterSubClasses(CBuilderObjectList &subs, inheritance_type itype);
	void FindAllSubClasses(CFCOClassBaseList &subs,bool includeSelf,bool includeAbstract, inheritance_type itype);
	void FindSubClasses(CBuilderObjectList &triangles, inheritance_type itype);
	void FindSubClasses();
	void FindEQClasses();
	void FindEQClasses(CBuilderObjectList &diamonds, bool leftIsProxy);
	void JoinEQClasses(CBuilderObjectList &eqs,bool leftIsProxy , CString commonName);
	void FindAttributes(CAttributeBuilderList &attrList);
	void FindAttributes();
	CAttributeBuilder *FindAttribute(CString nm);
	void OrderAttributes();
	CAttributeBuilder *FindLastAttribute();
	void AddConstraint(CConstraintBuilder *constraint);

	void DumpNamePosition(ofstream xml,int depth);
	void DumpGeneralPreferences(ofstream xml,int depth);
	void DumpHeader(ofstream xml,int depth,char *kind);
	void DumpDispName(ofstream xml,int depth);
	void DumpAttributes(ofstream xml,int depth);
	void DumpAttributeNames(ofstream xml,bool checkViewable = false);
	void DumpConstraints(ofstream xml,int depth);
	void DumpConnCardinalities(ofstream xml,int depth);
	void DumpFooter(ofstream xml,int depth,char *kind);
	void DumpIcon(ofstream xml, int depth);
	void DumpPortIcon(ofstream xml, int depth);
	void DumpDecorator(ofstream xml, int depth);
	void DumpHotspotEnabled(ofstream xml, int depth);
	void DumpTypeDisplayed(ofstream xml, int depth);
	void DumpSubtypeIcon( ofstream xml, int depth );
	void DumpInstanceIcon( ofstream xml, int depth );
	void DumpNameWrapNum( ofstream xml, int depth );
	void DumpNameEnabled( ofstream xml, int depth );
	bool IsRoot()	{ return root; }

	static void SetFlags(CFCOClassBaseList &roles,bool val);

	virtual void FindPotentialMembers(CFCOClassBaseList &potentialMembers) { }
	virtual void MatchMembers(CFCOClassBaseList &potentialMembers, CPointerItemList &list) { }
	virtual void DumpXML(ofstream xml,int depth);
};

class CFCOBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CFCOBuilder, CFCOClassBase)
public:
	virtual void Initialize();
};

class CModelBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CModelBuilder, CFCOClassBase)
public:
	virtual ~CModelBuilder();
protected:
	CAspectList aspects;
	CRoleList roles;

public:
//	void RegisterRole(CFCOClassBase *parent,
	void RegisterRole(CFCOClassBase *part,
					  CBuilderConnection *conn,
					  bool single);
	void FindRoles(CFCOClassBaseList &parts,
				   CBuilderConnection *conn );
//				   CFCOClassBase *parent);
	void FindRoles();
	bool HasRole(CRole *role) {return (roles.Find(role)!=NULL);};
	void FindAspects();
	CAspect *FindLastAspect();
	void OrderAspects();
	void FindMatchingPortRoles(CFCOClassBaseList &fcoClasses,CPointerItemList &ptrItems);
	void FindMatchingRoles(CFCOClassBaseList &fcoClasses,CPointerItemList &ptrItems,CRole *mainRole = 0);
	void DumpContainmentCardinalities(ofstream xml,int depth);
	void DumpTypeInfoDisplayed(ofstream xml, int depth);

	virtual void Initialize();
	virtual void DumpXML(ofstream xml,int depth);
};

class CAtomBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CAtomBuilder, CFCOClassBase)
public:
	virtual void DumpXML(ofstream xml,int depth);
};

class CReferenceBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CReferenceBuilder, CFCOClassBase)
protected:
	CFCOClassBaseList refs;
public:
	void FindRefs();
	virtual void DumpXML(ofstream xml,int depth);
	void GetReferredModels(CModelBuilderList &rmlist);
};

class CSetBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CSetBuilder, CFCOClassBase)
public:
	virtual ~CSetBuilder();
protected:
	CPointerItemList members;
public:
	void FindMembers();
	virtual void FindPotentialMembers(CFCOClassBaseList &potentialMembers);
	virtual void MatchMembers(CFCOClassBaseList &potentialMembers, CPointerItemList &list);
	virtual void DumpXML(ofstream xml,int depth);
};

class CConnectionBuilder : public CFCOClassBase
{
	DECLARE_CUSTOMATOM(CConnectionBuilder, CFCOClassBase)
public:
	virtual ~CConnectionBuilder();

public:
	CString srcRole;
	CString dstRole;
	CString srcCard;
	CString dstCard;
	CFCOClassBase *src;
	CFCOClassBase *dst;
protected:
	CFCOClassBaseList srcs;
	CFCOClassBaseList dsts;
	CPointerItemList srcPtrItems;
	CPointerItemList dstPtrItems;
	bool	bIsBidirectional;
public:
	bool FindSrcAndDst(bool hasBaseClasses);
	void BuildInstances();
	void DumpLabelSpec(ofstream xml,int depth,CString which,CString regLabel);
	void DumpXML(ofstream xml,int depth);
};

class CAttributeBuilder : public CUMLClassBase, public CBuilderAtom
{
	DECLARE_CUSTOMATOMBASE(CAttributeBuilder, CBuilderAtom)
protected:
	CString dispName;
	CString help;
	bool viewable;
	bool global;
	CRect loc;
public:
	bool IsGlobal()		{ return global; }
	bool IsViewable()	{ return viewable; }
	int Y()	 			{ return loc.CenterPoint().y; }
	int X()				{ return loc.CenterPoint().x; }
	virtual void Initialize();
	virtual void DumpXML(ofstream xml,int depth, const CString &path) {CUMLClassBase::DumpXML(xml, depth);};
};

class CBooleanAttributeBuilder : public CAttributeBuilder
{
	DECLARE_CUSTOMATOM(CBooleanAttributeBuilder, CAttributeBuilder)
public:
	virtual void DumpXML(ofstream xml,int depth, const CString &path);
};

class CFieldAttributeBuilder : public CAttributeBuilder
{
	DECLARE_CUSTOMATOM(CFieldAttributeBuilder, CAttributeBuilder)
public:
	virtual void DumpXML(ofstream xml,int depth, const CString &path);
};

class CEnumeratedAttributeBuilder : public CAttributeBuilder
{
	DECLARE_CUSTOMATOM(CEnumeratedAttributeBuilder, CAttributeBuilder)
public:
	virtual void DumpXML(ofstream xml,int depth, const CString &path);
};

class CRole {
public:
	CString baseName;
	CString name;
	CModelBuilder *parent;
	CBuilderConnection *conn;
	CFCOClassBase *part;
	int nMyOrder;
	static int nOrderNumber;
	bool linked;
	bool flag;
	CString cardinality;
public:
	CRole( );
	static void SetFlags(CRoleList &roles,bool val);
	static CRole *FindRoleByName(CRoleList &roles,CRole *start);
	static CRole *FindRoleByName(CRoleList &roles,CString name);
	void DumpXML(ofstream xml,int depth);
};

int CRole::nOrderNumber = 0;

class CPart {
public:
	CPart(CRole *r,CAspect *a) : role(r), aspect(a), kindaspect(a), isPrimary(true) {}
public:
	CRole *role;
	CAspect *aspect;
	CAspect *kindaspect;
	bool	isPrimary;
public:
	void DumpXML(ofstream xml,int depth);
};

class CAspect {
public:
	CAspect(CBuilderSet *s);
	virtual ~CAspect();

protected:
	CBuilderSet *set;
	CString name;
	CRect loc;
public:
	CBuilderObjectList parts;
	CPartList allParts;

	void FindSameAspects();
	void FindSameAspects(CBuilderObjectList &diamonds, bool leftIsProxy);
	void JoinSameAspects(CBuilderObjectList &sameaspects, bool leftIsProxy, CString commonName);
	CBuilderSet *GetSet()					{ return set; }
	CString GetName()						{ return name; }
	int Y()									{ return loc.CenterPoint().y; }
	CPart *FindPrimaryPart(CRole *role);
	void DumpXML(ofstream xml,int depth,CModelBuilder *owner);
	const CBuilderObject *GetBuilderObject() {return set;}
};

class CPointerItem {
public:
	CPointerItem() : object(0), port(0) {}
public:
	CRole *object;
	CRole *port;
public:
	static CPointerItem *FindEqualPointerItem(CPointerItemList &list,CRole *mainRole,CRole *role);
	void DumpXML(ofstream xml,int depth);

};

#endif // whole file