# <<NAME>> using GME dispatch based MgaComponent interface
# (c) 2001 ISIS, Vanderbilt University


import winerror
from win32com.server.exception import Exception

def gmePath():
    import win32con
    import win32api
    import os.path
    GMEAppPath = ('SOFTWARE', 'Microsoft', 'Windows', 'CurrentVersion', 'App Paths', 'GME.exe')
    base=win32con.HKEY_LOCAL_MACHINE
    for k in GMEAppPath:
        base = win32api.RegOpenKey(base,k,0,win32con.KEY_READ)
    return os.path.split(win32api.RegQueryValueEx(base, "Path")[0])[0]     

def makeModule(fname):
    import pythoncom
    import win32com.client
    from win32com.client import gencache, makepy
    tlb = pythoncom.LoadTypeLib(fname)
    sp = makepy.SimpleProgress(0)
    return gencache.MakeModuleForTypelibInterface(tlb, sp)

class <<ID>>:
    # COM-specific special attributes.
    _reg_clsid_ = "<<GUID>>"
    _reg_desc_ = "<<NAME>>"
    _reg_progid_ = "MGA.PythonInterpreter.<<ID>>"
    _reg_iconfile_ = "<<ICON>>"
    _public_methods_ = ['Enable', 'Initialize', 'Invoke', 'GetComponentParameter', 'SetComponentParameter', 'ObjectsInvokeEx', 'InvokeEx']
    _public_attrs_ = ['InteractiveMode',\
                      'ComponentName', 'ComponentType', 'Paradigm', 'ComponentProgID', 'Version']
    _paradigm_ = "<<PARADIGM>>"

    componentParameters = {}
        
    # Class methods.
    def __init__(self):
        # Get modules for python MGA interface definitions
        import win32com.client
        from win32com.client import gencache
        self.mga = gencache.GetModuleForProgID("Mga.MgaFCOs")
        self.meta = gencache.GetModuleForProgID("Mga.MgaMetaFCOs")
        self.mgautil = gencache.GetModuleForProgID("Mga.MgaRegistrar")

        self.InteractiveMode = 1
        self.ComponentName = <<ID>>._reg_desc_ 
        self.ComponentType = 1
        self.Paradigm = <<ID>>._paradigm_
        self.ComponentProgID = <<ID>>._reg_progid_
        self.Version = self.mgautil.constants.MgaInterfaceVersion_Current


    def Enable(self,b):
        pass

    def Initialize(self,project):
        pass
        
    def Invoke(self,project,sel,param):
        # This method is deprecated. Normally, GME never calls this
        pass

    def GetComponentParameter(self,paramname):
        if self.componentParameters.has_key(paramname):
            return self.componentParameters[paramname]
        else:
            return 0

    def SetComponentParameter(self,paramname,newval):
        self.componentParameters[paramname] = newval

    def ObjectsInvokeEx(self,project,currentobj,selectedobjs,param):
        # This method is currently not used by GME
        pass

    def InvokeEx(self,project,currentobj,selectedobjs,param):
    	project = self.mga.IMgaProject(project)
        terr = project.CreateTerritory(None, None, None)
        project.BeginTransaction(terr)

        # ...component work goes here

        # example:
        #msg = '<<ID>>:InvokeEx\n'
        #msg += '  project: ' + project.Name + '\n'
        #if currentobj != None:
        #    currentobj = self.mga.IMgaFCO(currentobj)
        #    msg += '  currentobj: ' + currentobj.Name + '\n'
        #else:
        #    msg += '  currentobj: None\n'
        #if selectedobjs != None:
        #    selectedobjs = self.mga.IMgaFCOs(selectedobjs)
        #    msg += '  selectedobjs.Count: ' + str(selectedobjs.Count) + '\n'
        #else:
        #    msg += '  selectedobjs: None\n'
        #msg += '  param: ' + str(param)
        #import win32ui
        #win32ui.MessageBox(msg)

    	project.CommitTransaction()

# Registration and unregistration functions-note that these are not methods!
GMECompRegPath=("SOFTWARE","Mga 2000","Components")

def Register():
    global GMECompRegPath
    import win32com.server.register
    import win32con
    import win32api
    
    win32com.server.register.UseCommandLine(<<ID>>)

    # Register as GME Interpreter
    base=win32con.HKEY_LOCAL_MACHINE
    keys,last=[],None
    for k in GMECompRegPath:
        base = win32api.RegOpenKey(base,k,0,win32con.KEY_ALL_ACCESS)
        keys.append(base)
    try:
        win32com.server.register.recurse_delete_key(<<ID>>._reg_progid_,base)
    except win32api.error, (code, fn, msg):
       if code != winerror.ERROR_FILE_NOT_FOUND:
          raise win32api.error, (code, fn, msg+path)
    hkey = win32api.RegCreateKey(base,<<ID>>._reg_progid_)
    keys.append(hkey)
    win32api.RegSetValueEx(hkey,"Description",None,win32con.REG_SZ,<<ID>>._reg_desc_)
    if hasattr(<<ID>>, '_reg_iconfile_'):
        if <<ID>>._reg_iconfile_ != "":
            win32api.RegSetValueEx(hkey,"Icon",None,win32con.REG_SZ,<<ID>>._reg_iconfile_)
    win32api.RegSetValueEx(hkey,"Type",None,win32con.REG_DWORD,1)
    base = hkey
    hkey = win32api.RegCreateKey(base,"Associated")
    plist = <<ID>>._paradigm_.split(',')
    for p in plist: 
        win32api.RegSetValueEx(hkey,p.strip(),None,win32con.REG_SZ,"")      
    keys.append(hkey)
    for k in keys:
        win32api.RegCloseKey(k)

    # get GME install directory from Registry and ensure pythonCOM support for needed Mga interfaces
    path = gmePath()
    makeModule(path + '\\bin\\Mga.dll')
    makeModule(path + '\\bin\\Meta.dll')
    makeModule(path + '\\bin\\MgaUtil.dll')

def UnRegister():
    import win32com.server.register
    from win32com.server.register import UnregisterServer
    import win32con
    import win32api

    UnregisterServer(<<ID>>()._reg_clsid_)
    # Unregister as GME interpreter
    base=win32con.HKEY_LOCAL_MACHINE
    keys,last=[],None
    for k in GMECompRegPath:
        base = win32api.RegOpenKey(base,k,0,win32con.KEY_ALL_ACCESS)
        keys.append(base)
    try:
        win32com.server.register.recurse_delete_key(<<ID>>._reg_progid_,base)
    except win32api.error, (code, fn, msg):
       if code != winerror.ERROR_FILE_NOT_FOUND:
          raise win32api.error, (code, fn, msg+path)
    for k in keys:
        win32api.RegCloseKey(k)
    print "Mga.PythonInterpreter.<<ID>> Class unregistered."

if __name__=='__main__':
    import sys
    if "/unreg" in sys.argv:
        UnRegister()
    else:
        Register()

